"""
Base classes for containers.
"""
from yadm.markers import AttributeNotSet
from yadm.fields.base import Field
from yadm.documents import DocumentItemMixin


class Container(DocumentItemMixin):
    """ Base class for containers.
    """
    def __init__(self, field, parent, value):
        self.__name__ = field.name
        self.__parent__ = parent
        self._field = field
        self._item_field = field.item_field
        self._data = value

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, item, value):
        self._data[item] = self._prepare_item(item, value)
        self._set_changed()

    def __delitem__(self, item):
        del self._data[item]
        self._set_changed()

    def __contains__(self, item):
        return item in self._data

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._data)

    def __eq__(self, other):
        if isinstance(other, Container):
            return self._data == other._data
        else:
            return self._data == other

    def _set_changed(self):
        self._field.set_parent_changed(self)

    def _prepare_item(self, item, value):
        return self._field.prepare_item(self, item, value)

    def _get_queryset(self):
        """ Return queryset for got data for this field.
        """
        if self.__db__ is None:
            raise RuntimeError('object not binded to database')

        qs = self.__db__.get_queryset(self.__document__.__class__)
        qs = qs.find({'_id': self.__document__.id})
        return qs.fields(self.__field_name__)

    def reload(self):
        """ Reload all object from database.
        """
        if len(list(self.__path__)) > 1:
            raise ValueError("can't reload deep objects: {}"
                             "".format(self.__field_name__))

        doc = self._get_queryset().read_primary().find_one()
        self._data = self.__get_value__(doc)._data


class ContainerField(Field):
    """ Base class for container fields.
    """
    container = Container

    def __init__(self, item_field=None, *, auto_create=True, **kwargs):
        super().__init__(**kwargs)

        if not isinstance(item_field, (Field, type(None))):
            raise TypeError("first argument must be field isinstance or None,"
                            " but {}".format(item_field))

        self.item_field = item_field
        self.auto_create = auto_create

    def get_default(self, document):
        if self.auto_create:
            return self.container(self, document, self.get_default_value())
        else:
            return AttributeNotSet

    def prepare_item(self, container, item, value):
        if self.item_field is not None:
            value = self.item_field.prepare_value(container, value)
            self._set_parent(container, item, value)
            return value
        else:
            raise NotImplementedError(
                "item_field is None, but prepare_item is not implemented")

    def prepare_value(self, document, value):
        # return self.container(self, document, value)
        raise NotImplementedError

    def get_default_value(self):
        raise NotImplementedError

    def to_mongo(self, document, value):
        # return value._data
        raise NotImplementedError

    def from_mongo(self, document, value):
        # return self.container(self, document, value)
        raise NotImplementedError

    def _set_parent(self, container, name, value):
        if isinstance(value, DocumentItemMixin):
            value.__parent__ = container
            value.__name__ = name

        return value



from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdoc'
    map = fields.MongoMapField()


def test_load(db):
    _id = db.db.testdoc.insert({'map': {'a': 1, 'b': '2', 'c': 'qwerty'}})
    doc = db.get_queryset(Doc).find_one(_id)

    assert hasattr(doc, 'map')
    assert isinstance(doc.map, fields.UnmutableMap)
    assert len(doc.map) == 3
    assert doc.map == {'a': 1, 'b': '2', 'c': 'qwerty'}


def test_setattr(db):
    _id = db.db.testdoc.insert({'map': {'a': 1}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map = {'b': 2}

    assert isinstance(doc.map, fields.UnmutableMap)
    assert doc.map == {'b': 2}


def test_setattr_save(db):
    _id = db.db.testdoc.insert({'map': {'a': 1}})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.map = {'b': 2}

    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['map'] == {'b': 2}
    assert doc.map == {'b': 2}



"""
Field for decimal numbers

.. code block: python

    class DocClass(Document):
        dec = DecimalField()

    doc = DocClass()
    doc.dec = Decimal('3.14')

    db.insert(doc)

This code save to MongoDB document:

.. code block: javascript

    {
        id: ObjectId('534272984c78591787e1a964'),
        dec: {i: 314, e: -2}
    }

"""
from decimal import Decimal, getcontext
from functools import reduce

from bson import Decimal128

from yadm.fields.base import Field, DefaultMixin, pass_null


class DecimalField(DefaultMixin, Field):
    """ Field for work with :class:`decimal.Decimal`.

    :param decimal.Context context: context for decimal operations
        (default: run :func:`decimal.getcontext` when need)
    :param decimal.Decimal default:

    TODO: context in copy()
    """
    _context = None

    def __init__(self, *, context=None, **kwargs):
        super().__init__(**kwargs)
        self.context = context

    @property
    def context(self):
        """ Context.

        :return: :class:`decimal.Context` for values
        """
        return self._context if self._context else getcontext()

    @context.setter
    def context(self, value):
        self._context = value

    def get_fake(self, document, faker, depth):
        return faker.pydecimal()

    @staticmethod
    def _integer_from_digits(digits):
        """ Make integer from digits.

        :param list digits: list of digits as integers
        :return: result integer

        [2, 4, 6, 7] => 2467
        """
        return reduce(lambda cur, acc: cur * 10 + acc, digits, 0)

    @pass_null
    def prepare_value(self, document, value):
        """ Cast value to :class:`decimal.Decimal`.
        """
        if isinstance(value, Decimal):
            return value
        elif isinstance(value, (str, int)):
            return Decimal(value, context=self.context)
        else:
            raise TypeError(value)

    @pass_null
    def to_mongo(self, document, value):
        sign, digits, exp = value.as_tuple()
        integer = self._integer_from_digits(digits)
        return {
            'i': -integer if sign else integer,
            'e': exp
        }

    @pass_null
    def from_mongo(self, document, value):
        if isinstance(value, dict):
            sign = value['i'] < 0  # False - positive, True - negative

            digits = []
            i = abs(value['i'])

            while i:
                i, d = divmod(i, 10)
                digits.append(d)

            digits.reverse()

            return Decimal((sign, digits, value['e']), context=self.context)

        elif isinstance(value, Decimal128):
            return value.to_decimal()

        elif isinstance(value, Decimal):
            return value

        else:
            raise TypeError(value)



import pytest

from pymongo.errors import BulkWriteError

from yadm import fields
from yadm.documents import Document
from yadm.bulk import Bulk


class Doc(Document):
    __collection__ = 'docs'

    i = fields.IntegerField()


@pytest.fixture
def index(db):
    return db.db.docs.ensure_index([('i', 1)], unique=True)


def test_create(db):
    bulk = db.bulk(Doc)
    assert isinstance(bulk, Bulk)


def test_insert_one(db):
    doc = Doc()
    doc.i = 1

    bulk = db.bulk(Doc)
    bulk.insert(doc)

    assert db.db.docs.count() == 0

    bulk.execute()

    assert bulk.result
    assert db.db.docs.count() == 1
    assert db.db.docs.find_one()['i'] == 1


def test_insert_many(db):
    bulk = db.bulk(Doc)

    for i in range(10):
        doc = Doc()
        doc.i = i
        bulk.insert(doc)

    assert db.db.docs.count() == 0
    bulk.execute()
    assert db.db.docs.count() == 10


def test_insert_type_error(db):
    class OtherDoc(Document):
        __collection__ = 'otherdocs'

    bulk = db.bulk(Doc)

    with pytest.raises(TypeError):
        bulk.insert(OtherDoc())


def test_context_manager(db):
    with db.bulk(Doc) as bulk:
        doc = Doc()
        doc.i = 1
        bulk.insert(doc)

    assert db.db.docs.count() == 1


def test_context_manager_error(db):
    with pytest.raises(RuntimeError):
        with db.bulk(Doc) as bulk:
            doc = Doc()
            doc.i = 1
            bulk.insert(doc)

            raise RuntimeError

    assert db.db.docs.count() == 0


def test_update_one(db):
    doc_one = db.insert(Doc(i=1))
    doc_two = db.insert(Doc(i=2))

    with db.bulk(Doc) as bulk:
        bulk.update_one(doc_one, set={'i': 7})
        bulk.update_one(doc_two, inc={'i': 11})

    assert db.reload(doc_one).i == 7
    assert db.reload(doc_two).i == 13


def test_find_update(db):
    doc_one = db.insert(Doc(i=1))
    doc_two = db.insert(Doc(i=2))
    doc_three = db.insert(Doc(i=3))

    with db.bulk(Doc) as bulk:
        bulk.find({'i': {'$gte': 2}}).update(inc={'i': 11})

    assert db.reload(doc_one).i == 1
    assert db.reload(doc_two).i == 13
    assert db.reload(doc_three).i == 14


def test_find_upsert_update(db):
    with db.bulk(Doc) as bulk:
        bulk.find({'i': 10}).upsert().update(inc={'i': 3})

    assert len(db(Doc)) == 1
    assert db(Doc).find_one().i == 13


def test_find_remove(db):
    for i in range(20):
        db.insert(Doc(i=i))

    with db.bulk(Doc) as bulk:
        bulk.find({'i': 13}).remove()
        bulk.find({'i': {'$lt': 7}}).remove()
        bulk.find({'i': {'$gte': 15}}).remove()

    assert len(db(Doc)) == 7
    assert set(db(Doc).distinct('i')) == {7, 8, 9, 10, 11, 12, 14}


def test_find_mixed(db):
    for i in range(5):
        db.insert(Doc(i=i))

    with db.bulk(Doc, ordered=True) as bulk:
        for i in range(5, 10):
            bulk.insert(Doc(i=i))

        assert len(db(Doc)) == 5
        bulk.find({'i': {'$gte': 7}}).update(inc={'i': 10})
        assert len(db(Doc)) == 5

        bulk.find({'i': {'$lt': 4}}).remove()
        assert len(db(Doc)) == 5

    assert len(db(Doc)) == 6
    assert set(db(Doc).distinct('i')) == {4, 5, 6, 17, 18, 19}


def test_result(db):
    with db.bulk(Doc) as bulk:
        doc = Doc()
        doc.i = 1
        bulk.insert(doc)

    assert bulk.result.n_inserted == 1


def test_result_write_error(db, index):
    with db.bulk(Doc, raise_on_errors=False) as bulk:
        doc = Doc()
        doc.i = 1
        bulk.insert(doc)

        doc = Doc()
        doc.i = 2
        bulk.insert(doc)

        doc = Doc()
        doc.i = 1
        bulk.insert(doc)

    assert not bulk.result
    assert db.db.docs.count() == 2
    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1


def test_result_write_error_raise(db, index):
    with pytest.raises(BulkWriteError):
        with db.bulk(Doc) as bulk:
            doc = Doc()
            doc.i = 1
            bulk.insert(doc)

            doc = Doc()
            doc.i = 2
            bulk.insert(doc)

            doc = Doc()
            doc.i = 1
            bulk.insert(doc)

    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1


def test_result_write_error_ordered(db, index):
    with db.bulk(Doc, ordered=True, raise_on_errors=False) as bulk:
        for i in range(10):
            doc = Doc()
            doc.i = i
            bulk.insert(doc)

            doc = Doc()
            doc.i = 1
            bulk.insert(doc)

    assert db.db.docs.count() == 2
    assert bulk.result.n_inserted == 2
    assert len(bulk.result.write_errors) == 1
    assert bulk.result.write_errors[0].document.i == 1


def test_repr(db):
    assert 'Bulk' in repr(db.bulk(Doc))


def test_find_repr(db):
    assert 'BulkQuery' in repr(db.bulk(Doc).find({'i': 1}))
    assert ': 13' in repr(db.bulk(Doc).find({'i': 13}))
    assert 'upsertable' in repr(db.bulk(Doc).find({'i': 13}).upsert())



import pytest

from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdoc'
    li = fields.ListField(fields.IntegerField())


def test_default():
    doc = Doc()
    assert isinstance(doc.li, fields.list.List)
    assert not doc.li
    assert len(doc.li) == 0
    assert doc.li._data == []
    assert doc.li == []


def test_default_no_auto():
    class Doc(Document):
        __collection__ = 'testdoc'
        li = fields.ListField(fields.IntegerField(), auto_create=False)

    doc = Doc()
    assert not hasattr(doc, 'li')


def test_get(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)

    assert doc.li
    assert len(doc.li) == 3
    assert doc.li._data == [1, 2, 3]
    assert list(doc.li) == [1, 2, 3]
    assert doc.li[1] == 2


def test_append(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.append(4)
    assert doc.li == [1, 2, 3, 4]


def test_append_valueerror():
    doc = Doc()
    with pytest.raises(ValueError):
        doc.li.append('not a number')


def test_append_save(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.append(4)
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['li'] == [1, 2, 3, 4]


def test_remove(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.remove(2)

    assert doc.li == [1, 3]


def test_remove_save(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.remove(2)
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['li'] == [1, 3]


def test_push(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.push(4)

    assert doc.li == [1, 2, 3, 4]

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['li'] == [1, 2, 3, 4]


def test_push_valueerror():
    doc = Doc()
    with pytest.raises(ValueError):
        doc.li.push('not a number')


def test_pull(db):
    _id = db.db.testdoc.insert({'li': [1, 2, 3]})
    doc = db.get_queryset(Doc).find_one(_id)
    doc.li.pull(2)

    assert doc.li == [1, 3]

    data = db.db.testdoc.find_one({'_id': _id})
    assert data['li'] == [1, 3]



from bson import ObjectId

from yadm.documents import Document, EmbeddedDocument
from yadm.testing import create_fake
from yadm.fields import (
    BooleanField, StringField, IntegerField, ObjectIdField, EmailField,
    EmbeddedDocumentField, ReferenceField,
)


class SimpleDoc(Document):
    __collection__ = 'testdocs'

    oid = ObjectIdField()
    b = BooleanField()
    s = StringField()
    i = IntegerField()
    e = EmailField()
    email = StringField()


class EmbeddedDoc(EmbeddedDocument):
    __collection__ = 'testdocs'

    name = StringField()


def test_simple(db):
    doc = create_fake(SimpleDoc)

    assert doc.__db__ is None

    assert hasattr(doc, 'oid')
    assert hasattr(doc, 'b')
    assert hasattr(doc, 's')
    assert hasattr(doc, 'i')
    assert hasattr(doc, 'e')
    assert hasattr(doc, 'email')

    assert isinstance(doc.oid, ObjectId)
    assert isinstance(doc.b, bool)
    assert isinstance(doc.s, str)
    assert isinstance(doc.i, int)
    assert isinstance(doc.e, str)
    assert isinstance(doc.email, str)

    assert len(doc.s) > 0
    assert '@' in doc.e
    assert '@' in doc.email

    assert db(SimpleDoc).count() == 0


def test_simple_save(db):
    doc = create_fake(SimpleDoc, __db__=db)

    assert doc.__db__ is db

    assert hasattr(doc, 'oid')
    assert hasattr(doc, 'b')
    assert hasattr(doc, 's')
    assert hasattr(doc, 'i')

    assert isinstance(doc.oid, ObjectId)
    assert isinstance(doc.b, bool)
    assert isinstance(doc.s, str)
    assert isinstance(doc.i, int)

    assert len(doc.s) > 0

    assert db(SimpleDoc).count() == 1
    assert db(SimpleDoc).find_one(doc.id)


class WithEmbeddedDoc(Document):
    __collection__ = 'testdocs'

    emb = EmbeddedDocumentField(EmbeddedDoc)


def test_embedded(db):
    doc = create_fake(WithEmbeddedDoc)

    assert hasattr(doc, 'emb')
    assert isinstance(doc.emb, EmbeddedDoc)
    assert isinstance(doc.emb.name, str)


def test_embedded_depth_limit(db):
    doc = create_fake(WithEmbeddedDoc, __depth__=0)

    assert hasattr(doc, 'emb')
    assert not hasattr(doc.emb, 'name')


class WithReferenceDoc(Document):
    __collection__ = 'with_ref_testdocs'

    ref = ReferenceField('tests.test_testing.SimpleDoc')


def test_reference(db):
    doc = create_fake(WithReferenceDoc)

    assert hasattr(doc, 'ref')
    assert hasattr(doc.ref, 's')
    assert isinstance(doc.ref.s, str)

    assert db(WithReferenceDoc).count() == 0
    assert db(SimpleDoc).count() == 0


def test_reference_save(db):
    doc = create_fake(WithReferenceDoc, __db__=db)

    assert hasattr(doc, 'ref')
    assert hasattr(doc.ref, 's')
    assert isinstance(doc.ref.s, str)

    assert db(WithReferenceDoc).count() == 1
    assert db(SimpleDoc).count() == 1


class WithReferenceCircleDoc(Document):
    __collection__ = 'testdocs'

    self = ReferenceField('tests.test_testing.WithReferenceCircleDoc')


def test_reference_circle(db):
    doc = create_fake(WithReferenceCircleDoc, __db__=db, __depth__=2)

    assert db(WithReferenceCircleDoc).count() == 3

    assert hasattr(doc, 'self')
    assert hasattr(doc.self, 'self')
    assert not hasattr(doc.self.self, 'self')



"""
Base classes for build database fields.
"""
import functools

from yadm.markers import AttributeNotSet, NotLoaded


class NotLoadedError(Exception):
    """ Raise if value marked as not loaded.

    .. code:: python

        doc = db(Doc).fields('a').find_one()
        try:
            doc.b
        except NotLoadedError:
            print("raised!")
    """


def pass_null(method):
    @functools.wraps(method)
    def wrapper(self, document, value):
        if value is None:
            return None
        else:
            return method(self, document, value)

    return wrapper


class FieldDescriptor:
    """ Base desctiptor for fields.

    .. py:attribute:: name

    Name of field

    .. py:attribute:: field

    Field instance for this desctiptor

    """
    def __init__(self, name, field):
        self.name = name
        self.field = field

    def __repr__(self):  # pragma: no cover
        class_name = type(self).__name__
        document_class_name = type(self.field.document_class).__name__
        return '<{} "{}.{}">'.format(class_name, document_class_name, self.name)

    def __get__(self, instance, owner):
        """ Get python value from document.

        1. Lookup in __changed__;
        2. Lookup in __cache__;
        3. Lookup in __raw__:

            - if AttributeNotSet -- call Field.get_if_attribute_not_set;
            - if NotLoaded -- call Field.get_if_not_loaded;
            - call Field.from_mongo;
            - set __name__ and __parent__
            - save to __cache__

        4. Call Field.get_default;
        5. If AttributeNotSet -- call Field.get_if_attribute_not_set;
        6. Return value.
        """
        name = self.name

        if instance is None:
            return self.field

        elif name in instance.__changed__:
            value = instance.__changed__[name]

        elif name in instance.__cache__:
            value = instance.__cache__[name]

        elif name in instance.__raw__:
            raw = instance.__raw__[name]

            if raw is AttributeNotSet:
                return self.field.get_if_attribute_not_set(instance)
            if raw is NotLoaded:
                return self.field.get_if_not_loaded(instance)

            value = self.field.from_mongo(instance, raw)

            from yadm.documents import DocumentItemMixin
            if isinstance(value, DocumentItemMixin):
                value.__name__ = self.field.name
                value.__parent__ = instance

            instance.__cache__[name] = value

        else:
            value = self.field.get_default(instance)
            instance.__changed__[name] = value

        if value is AttributeNotSet:
            return self.field.get_if_attribute_not_set(instance)

        return value

    def __set__(self, instance, value):
        """ Set value to document.

        1. Call Field.prepare_value for cast value;
        2. Save in Document.__changed__;
        3. Call Field.set_parent_changed.
        """
        if not isinstance(instance, type):
            value = self.field.prepare_value(instance, value)

            name = self.name

            if name in instance.__changed__:
                value_old = instance.__changed__[name]
            elif name in instance.__cache__:
                value_old = instance.__cache__[name]
            elif name in instance.__raw__:
                try:
                    value_old = getattr(instance, name, AttributeNotSet)
                except NotLoadedError:
                    value_old = NotLoaded
            else:
                value_old = AttributeNotSet

            if value != value_old:
                instance.__changed__[name] = value
                self.field.set_parent_changed(instance)

        else:
            raise TypeError("can't set field directly")  # pragma: no cover

    def __delete__(self, instance):
        """ Mark document's key as not set.
        """
        if not isinstance(instance, type):
            setattr(instance, self.name, AttributeNotSet)


class Field:
    """ Base field for all database fields.

    :param bool smart_null:

        If it `True`, access to not exists fields return
        `None` instead `AttributeError` exception.
        You will not be able to distinguish null value from
        not exist. Use with care.

    .. py:attribute:: descriptor_class

        Class of desctiptor for work with field

    .. py:attribute:: document_class

        Class of document.
        Set in :py:meth:`contribute_to_class`.

    .. py:attribute:: name

        Name of field in document.
        Set in :py:meth:`contribute_to_class`.
    """
    descriptor_class = FieldDescriptor
    smart_null = False
    document_class = None
    name = None

    def __init__(self, smart_null=False):
        self.smart_null = smart_null

    def __repr__(self):
        class_name = type(self).__name__
        doc_class_name = self.document_class and self.document_class.__name__
        return '<{} "{}.{}">'.format(class_name, doc_class_name, self.name)

    def contribute_to_class(self, document_class, name):
        """ Add field for document_class.

        :param MetaDocument document_class: document class for add
        """
        self.name = name
        self.document_class = document_class
        self.document_class.__fields__[name] = self
        setattr(document_class, name, self.descriptor_class(name, self))

    def set_parent_changed(self, instance):
        from yadm.documents import DocumentItemMixin
        if (isinstance(instance, DocumentItemMixin) and
                instance.__parent__ is not None):

            first = list(instance.__path__)[-1]
            first_name = first.__name__
            instance.__document__.__changed__[first_name] = first
            # TODO: update __changed__ for full path

    def copy(self):
        """ Return copy of field.
        """
        return self.__class__(smart_null=self.smart_null)

    def get_default(self, document):
        """ Return default value.
        """
        return AttributeNotSet

    def get_if_not_loaded(self, document):
        """ Call if field data marked as not loaded.
        """
        raise NotLoadedError(self, document)

    def get_if_attribute_not_set(self, document):
        """ Call if key not exist in document.
        """
        raise AttributeError("{!r} document has no attribute {!r}"
                             "".format(document.__class__.__name__, self.name))

    def get_fake(self, document, faker, deep):  # pragma: no cover
        """ Return fake data for testing.
        """
        return self.get_default(document)

    def prepare_value(self, document, value):  # pragma: no cover
        """ The method is called when value is assigned for the attribute.

        :param BaseDocument document: document
        :param value: raw value
        :return: prepared value

        It must be accept `value` argument and return
        processed (e.g. casted) analog. Also it is called
        once for the default value.
        """
        return value

    def to_mongo(self, document, value):  # pragma: no cover
        """ Convert python value to mongo value.

        :param BaseDocument document: document
        :param value: python value
        :return: mongo value
        """
        return value

    def from_mongo(self, document, value):  # pragma: no cover
        """ Convert mongo value to python value.

        :param BaseDocument document: document
        :param value: mongo value
        :return: python value
        """
        return value


class DefaultMixin:
    default = AttributeNotSet

    def __init__(self, *args, default=AttributeNotSet, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = default

    def get_default(self, document):
        return self.default

    def copy(self):
        """ Return copy of field.
        """
        return self.__class__(smart_null=self.smart_null, default=self.default)



"""
Map.

.. code-block:: python

    class Doc(Document):
        __collection__ = 'docs'
        map = fields.MapField(fields.IntegerField)

    doc = Doc()
    doc.map['a'] = 1
    doc.map['b'] = 2
    assert doc.map == {'a': 1, 'b': 2}

    db.insert(doc)
    doc = db.get_queryset(Doc).find_one(doc.id)  # reload

    doc.map['c'] = 3  # do not save
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).find_one(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    del doc.map['b']  # do not save too
    assert doc.map == {'a': 1}
    doc = db.get_queryset(Doc).find_one(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    doc.map.set('d', 3)  # $set query
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).find_one(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}

    doc.map.unset('d', 3)  # $unset query
    assert doc.map == {'a': 1, 'b': 2}
    doc = db.get_queryset(Doc).find_one(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

"""
from collections import abc

from yadm.markers import AttributeNotSet
from yadm.fields.base import pass_null
from yadm.fields.containers import (
    Container,
    ContainerField,
)


class Map(Container, abc.MutableMapping):
    """ Map.
    """
    def set(self, key, value, reload=True):
        """ Set key directly in database.

        :param key: key
        :param value: value for `$set`

        See `$set` in MongoDB's `set`.
        """
        value = self._prepare_item(key, value)
        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$set': {fn: value}}, multi=False)
        self._data[key] = value

        if reload:
            self.reload()

    def unset(self, key, reload=True):
        """ Unset key directly in database.

        :param key: key

        See `$unset` in MongoDB's `unset`.
        """
        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$unset': {fn: True}}, multi=False)
        del self._data[key]

        if reload:
            self.reload()


class MapField(ContainerField):
    """ Field for maps.
    """
    container = Map

    def get_default_value(self):
        return {}

    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return AttributeNotSet

        pi = self.prepare_item
        container = self.container(self, document, {})

        if isinstance(value, dict):
            items = value.items()
        else:
            items = value

        container._data.update((k, pi(container, k, i)) for k, i in items)
        return container

    @pass_null
    def to_mongo(self, document, value):
        tm = self.item_field.to_mongo
        return {k: tm(value, i) for k, i in value.items()}

    @pass_null
    def from_mongo(self, document, value):
        fm = self.item_field.from_mongo
        sp = self._set_parent

        container = self.container(self, document, {})
        g = ((k, sp(container, k, fm(container, i))) for k, i in value.items())
        container._data.update(g)
        return container


class MapCustomKeys(Map):
    def __init__(self, field, parent, value):
        k2s = field.key_to_str
        pi = field.prepare_item
        value = {k2s(k): pi(self, k, v) for k, v in value.items()}

        super().__init__(field, parent, value)
        self.key_factory = field.key_factory
        self.key_to_str = field.key_to_str

    def __iter__(self):
        key_factory = self.key_factory
        return (key_factory(k) for k in super().__iter__())

    def __getitem__(self, item):
        return super().__getitem__(self.key_to_str(item))

    def __setitem__(self, item, value):
        super().__setitem__(self.key_to_str(item), value)

    def __delitem__(self, item):
        super().__delitem__(self.key_to_str(item))

    def __contains__(self, item):
        return self.key_to_str(item) in self._data

    def __eq__(self, other):
        if other is None or other is AttributeNotSet:
            return False
        elif isinstance(other, dict):
            other = self.__class__(self._field, self.__parent__, other)

        return self._data == other._data

    def set(self, key, value, reload=True):
        return super().set(self.key_to_str(key), value, reload)

    def unset(self, key, value, reload=True):
        return super().unset(self.key_to_str(key), value, reload)


class MapCustomKeysField(MapField):
    """ Field for maps with custom key type.

    :param field item_field:
    :param func key_factory: function, who return thue key
        from raw string key
    :param func key_to_str:
    :param bool auto_create:
    """
    container = MapCustomKeys

    def __init__(self, item_field, key_factory, *,
                 key_to_str=str,
                 auto_create=True,
                 **kwargs):
        super().__init__(item_field, auto_create=auto_create, **kwargs)
        self.key_factory = key_factory
        self.key_to_str = key_to_str

    @pass_null
    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return AttributeNotSet

        pi = self.prepare_item
        k2s = self.key_to_str

        container = self.container(self, document, {})

        if isinstance(value, dict):
            items = value.items()
        else:
            items = value

        container._data.update((k2s(k), pi(container, k, i)) for k, i in items)
        return container

    @pass_null
    def to_mongo(self, document, value):
        tm = self.item_field.to_mongo
        k2s = self.key_to_str
        return {k2s(k): tm(value, i) for k, i in value.items()}



import random

import pytest

import pymongo
from bson import ObjectId

from yadm.documents import Document
from yadm import fields
from yadm.markers import NotLoaded
from yadm.results import UpdateResult, RemoveResult
from yadm.queryset import QuerySet, NotFoundError


class Doc(Document):
    __collection__ = 'testdocs'
    i = fields.IntegerField()
    s = fields.StringField()


@pytest.fixture
def qs(db):
    for n in range(10):
        db.db.testdocs.insert({
            'i': n,
            's': 'str({})'.format(n),
        })

    return db.get_queryset(Doc)


def test_count(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert qs.count() == 4


def test_len(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert len(qs) == 4


@pytest.mark.parametrize('slice, result', [
    (slice(None, 2), [6, 7]),
    (slice(2, None), [8, 9]),
    (slice(1, 3), [7, 8]),
    (slice(100, 300), []),
    (slice(None, -3), TypeError),
    (slice(-3, None), TypeError),
    (slice(-3, -1), TypeError),
    (slice(None, None, 2), TypeError),
])
def test_slice(qs, slice, result):
    if isinstance(result, type) and issubclass(result, Exception):
        with pytest.raises(result):
            qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))[slice]
    else:
        qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))[slice]
        assert [d.i for d in qs] == result


def test_get_one(qs):
    qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))
    doc = qs[1]
    assert doc.i == 7


def test_get_one__index_error(qs):
    qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))
    with pytest.raises(IndexError):
        qs[100]


def test_find_one(qs):
    doc = qs.find_one({'i': 7})
    assert isinstance(doc, Doc)
    assert hasattr(doc, 'i')
    assert doc.i == 7


def test_find_exc(qs):
    class SimpleException(Exception):
        pass

    with pytest.raises(SimpleException):
        qs.find_one({'i': 100500}, exc=SimpleException)


def test_find(qs):
    qs = qs.find({'i': {'$gte': 6}})
    assert len([d for d in qs]) == 4
    assert {d.i for d in qs} == {6, 7, 8, 9}


def test_find_with_collisium(qs):
    qs = qs.find({'i': {'$gt': 4}})
    qs = qs.find({'i': {'$lt': 6}})
    assert qs.count() == 1
    assert qs[0].i == 5


def test_update(db, qs):
    result = qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}})

    assert isinstance(result, UpdateResult)
    assert result
    assert result.matched == result.modified == int(result) == 4
    assert result.upserted == 0

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 4

    for doc in db.db.testdocs.find({'i': {'$lt': 6}}):
        assert doc['s'] != 'test'
        assert doc['s'].startswith('str(')

    for doc in db.db.testdocs.find({'i': {'$gte': 6}}):
        assert doc['s'] == 'test'


def test_update_not_multi(db, qs):
    result = qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}}, multi=False)

    assert isinstance(result, UpdateResult)
    assert result
    assert result.matched == result.modified == int(result) == 1
    assert result.upserted == 0

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1


def test_find_and_modify(db, qs):
    doc = qs({'i': 6}).find_and_modify({'$set': {'s': 'test'}})

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'str(6)'


def test_find_and_modify_new(db, qs):
    doc = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, new=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'test'


def test_find_and_modify_full_response(db, qs):
    result = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, full_response=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert 'value' in result
    assert isinstance(result['value'], Document)
    assert result['value'].i == 6
    assert result['value'].s == 'str(6)'


def test_find_and_modify_full_response_new(db, qs):
    result = qs({'i': 6}).find_and_modify(
        {'$set': {'s': 'test'}}, full_response=True, new=True)

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert 'value' in result
    assert isinstance(result['value'], Document)
    assert result['value'].i == 6
    assert result['value'].s == 'test'


def test_find_and_modify_sort(db, qs):
    qs = qs({'i': {'$lte': 6, '$gte': 4}}).sort(('s', -1))
    doc = qs.find_and_modify({'$set': {'s': 'test'}})

    assert db.db.testdocs.count() == 10
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(10))
    assert db.db.testdocs.find({'s': 'test'}).count() == 1
    assert db.db.testdocs.find_one({'i': 6})['s'] == 'test'

    assert isinstance(doc, Document)
    assert doc.i == 6
    assert doc.s == 'str(6)'


def test_find_and_modify_not_found(qs):
    ret = qs({'i': 13}).find_and_modify({'$set': {'s': 'test'}})
    assert ret is None


def test_remove(db, qs):
    result = qs.find({'i': {'$gte': 6}}).remove()

    assert isinstance(result, RemoveResult)
    assert result
    assert result.removed == int(result) == 4

    assert len([d for d in qs]) == 6
    assert {d.i for d in qs} == set(range(6))

    assert db.db.testdocs.count() == 6
    assert {d['i'] for d in db.db.testdocs.find()} == set(range(6))


def test_remove__one(db, qs):
    result = qs.find({'i': {'$gte': 6}}).remove(multi=False)

    assert isinstance(result, RemoveResult)
    assert result
    assert result.removed == int(result) == 1

    assert len([d for d in qs]) == 9
    assert db.db.testdocs.count() == 9


def test_sort(qs):
    qs = qs.find({'i': {'$gte': 6}}).sort(('i', -1))
    assert [d.i for d in qs] == [9, 8, 7, 6]

    qs = qs.find({'i': {'$gte': 6}}).sort(('i', 1))
    assert [d.i for d in qs] == [6, 7, 8, 9]


def test_fields(qs):
    doc = qs.fields('s').find_one({'i': 3})

    assert doc.s, 'str(3)'
    assert 'i' in doc.__raw__
    assert doc.__raw__['i'] is NotLoaded

    with pytest.raises(fields.NotLoadedError):
        doc.i


def test_fields_all(qs):
    doc = qs.fields('s').fields_all().find_one({'i': 3})
    assert 'i' in doc.__data__
    assert doc.__data__['i'] is 3
    assert doc.s == 'str(3)'
    assert doc.i == 3


def test_read_preference(db):
    sp = pymongo.read_preferences.ReadPreference.SECONDARY_PREFERRED
    qs = db(Doc).read_preference(sp)
    assert qs._collection.read_preference == sp


def test_with_id(db, qs):
    _id = db.db.testdocs.find_one({'i': 4}, {'_id': True})['_id']
    doc = qs.with_id(_id)
    assert doc.s == 'str(4)'
    assert doc.i == 4


def test_contains(qs):
    doc = qs.find_one({'i': 0})
    assert doc in qs
    assert doc not in qs.find({'i': {'$ne': 0}})


def test_ids(db):
    qs = db(Doc)
    assert not list(qs.ids())

    ids = set()
    ids.add(db.insert(Doc(i=3)))
    ids.add(db.insert(Doc(i=4)))
    ids.add(db.insert(Doc(i=10)))

    assert len(list(qs.ids())) == 3
    assert set(qs.ids()) == ids


def test_distinct(db, qs):
    res = qs.distinct('i')
    assert isinstance(res, list)
    assert len(res) == len(set(res))

    db.insert(Doc(i=3))
    db.insert(Doc(i=4))
    db.insert(Doc(i=10))

    res = qs.distinct('i')
    assert len(res) == len(set(res))


def test_bulk(qs):
    bulk = qs.find({'i': {'$gte': 6}}).bulk()
    assert isinstance(bulk, dict)
    assert len(bulk), 4
    assert isinstance(list(bulk)[0], ObjectId)
    assert isinstance(list(bulk.values())[0], Doc)
    _id = list(bulk)[0]
    assert bulk[_id].id == _id
    assert {d.i for d in bulk.values()} == {6, 7, 8, 9}


@pytest.mark.parametrize('preferred, rp', [
    (False, pymongo.read_preferences.Primary()),
    (True, pymongo.read_preferences.PrimaryPreferred()),
])
def test_read_primary(qs, preferred, rp):
    assert not qs._collection_params
    qs = qs.read_primary(preferred=preferred)
    assert qs._collection_params['read_preference'] == rp
    assert len(qs)  # try to execute query


def test_batch_size__set():
    qs_one = QuerySet(None, Doc)
    assert qs_one._batch_size is None

    qs_two = qs_one.batch_size(10)
    assert qs_two._batch_size == 10
    assert qs_one._batch_size is None

    qs_three = qs_two.batch_size(None)
    assert qs_three._batch_size is None
    assert qs_two._batch_size == 10
    assert qs_one._batch_size is None


def test_batch_size__get(qs):
    n = 0
    for doc in qs.batch_size(10):
        n += 1

    assert n > 0


def test_default_projection(db, qs):
    class ProjectedDoc(Doc):
        __default_projection__ = {'s': False}

    qs = db.get_queryset(ProjectedDoc)
    doc = qs.find_one()

    assert isinstance(doc.i, int)

    with pytest.raises(fields.base.NotLoadedError):
        doc.s


class TestFindIn:
    @pytest.fixture(autouse=True)
    def ids(self, qs):
        ids = [doc.id for doc in qs]
        random.shuffle(ids)
        return ids

    def test_simple(self, qs, ids):
        result = [getattr(doc, 'id', doc)
                  for doc in qs.find_in(ids)]

        assert result == ids

    def test_skip(self, qs, ids):
        ids.append('NotExistId')
        result = [doc.id for doc in qs.find_in(ids)]
        assert result == ids[:-1]

    def test_none(self, qs, ids):
        index = random.randrange(len(ids))
        ids[index] = 'NotExistId'
        result = [getattr(doc, 'id', doc)
                  for doc in qs.find_in(ids, not_found='none')]

        ids[index] = None
        assert result == ids

    def test_exception(self, qs, ids):
        [getattr(doc, 'id', doc)
         for doc in qs.find_in(ids, not_found='error')]

        with pytest.raises(NotFoundError):
            ids.append('NotExistId')
            [getattr(doc, 'id', doc)
             for doc in qs.find_in(ids, not_found='error')]

    def test_copy_id(self, qs, ids):
        ids.append(ids[0])
        result = [getattr(doc, 'id', doc)
                  for doc in qs.find_in(ids)]

        assert result == ids



import pytest
from bson import ObjectId

from yadm import fields
from yadm.documents import Document, EmbeddedDocument
from yadm.testing import create_fake


class EDoc(EmbeddedDocument):
    id = fields.ObjectIdField(default_gen=True)
    i = fields.IntegerField()
    s = fields.StringField(default='default')
    e = fields.EmbeddedDocumentField('self')


class Doc(Document):
    __collection__ = 'testdoc'
    e = fields.EmbeddedDocumentField(EDoc, auto_create=False)


class DocAuto(Document):
    __collection__ = 'testdoc'
    e = fields.EmbeddedDocumentField(EDoc)


def test_default():
    doc = Doc()
    assert not hasattr(doc, 'e')


def test_default_auto():
    doc = DocAuto()
    assert hasattr(doc, 'e')
    assert isinstance(doc.e, EDoc)


def test_default_auto_not_save_empty(db):
    class ETD(EmbeddedDocument):
        i = fields.IntegerField()

    class TD(Document):
        __collection__ = 'testdoc'
        e = fields.EmbeddedDocumentField(ETD)

    doc = db.save(TD())
    raw = db.db.testdoc.find_one(doc.id)
    assert 'e' not in raw
    assert set(raw) == {'_id'}

    doc.e.i = 13
    db.save(doc)
    raw = db.db.testdoc.find_one(doc.id)
    assert 'e' in raw
    assert raw == {'_id': doc.id, 'e': {'i': 13}}


@pytest.mark.parametrize('raw', [
    {'i': 13},
    {'i': 13, 's': 'defined'},
    {'s': 'defined'},
    {},
    None,
])
def test_get(db, raw):
    if raw is not None:
        _id = db.db.testdoc.insert({'e': raw})
    else:
        _id = db.db.testdoc.insert({})

    doc = db.get_queryset(Doc).find_one(_id)

    if raw is None:
        assert not hasattr(doc, 'e')
    else:
        assert hasattr(doc, 'e')
        assert isinstance(doc.e, EDoc)

        if 'i' in raw:
            assert hasattr(doc.e, 'i')
            assert isinstance(doc.e.i, int)
            assert doc.e.i == raw['i']
        else:
            assert not hasattr(doc.e, 'i')

        if 's' in raw:
            assert hasattr(doc.e, 's')
            assert isinstance(doc.e.s, str)
            assert doc.e.s == raw['s']
        else:
            assert not hasattr(doc.e, 's')


@pytest.mark.parametrize('raw', [
    {'i': 13},
    {'i': 13, 's': 'defined'},
    {'s': 'defined'},
    {},
    None,
])
def test_get_deeper(db, raw):
    if raw is not None:
        _id = db.db.testdoc.insert({'e': {'e': raw}})
    else:
        _id = db.db.testdoc.insert({'e': {}})

    doc = db.get_queryset(Doc).find_one(_id)

    if raw is None:
        assert hasattr(doc.e, 'e')
        assert not hasattr(doc.e.e, 'i')
        assert doc.e.e.s == 'default'
    else:
        assert hasattr(doc.e, 'e')
        assert isinstance(doc.e.e, EDoc)

        if 'i' in raw:
            assert hasattr(doc.e.e, 'i')
            assert isinstance(doc.e.e.i, int)
            assert doc.e.e.i == raw['i']
        else:
            assert not hasattr(doc.e.e, 'i')

        if 's' in raw:
            assert hasattr(doc.e.e, 's')
            assert isinstance(doc.e.e.s, str)
            assert doc.e.e.s == raw['s']
        else:
            assert not hasattr(doc.e.e, 's')


def test_set():
    doc = Doc()
    doc.e = EDoc()

    assert hasattr(doc, 'e')
    assert isinstance(doc.e, EDoc)
    assert not hasattr(doc.e, 'i')

    doc.e.i = 13
    assert hasattr(doc.e, 'i')
    assert isinstance(doc.e.i, int)
    assert doc.e.i == 13


def test_set_typeerror():
    class FailEDoc(EmbeddedDocument):
        s = fields.StringField

    doc = Doc()

    with pytest.raises(TypeError):
        doc.e = FailEDoc()


def test_set_insert_wo_set(db):
    doc = Doc()
    db.insert(doc)

    data = db.db.testdoc.find_one({'_id': doc.id})
    assert set(data) == {'_id'}
    assert data['_id'] == doc.id


def test_set_insert(db):
    doc = Doc()
    doc.e = EDoc()
    doc.e.i = 13
    db.insert(doc)

    data = db.db.testdoc.find_one({'_id': doc.id})
    assert set(data) == {'_id', 'e'}
    assert data['_id'] == doc.id
    assert data['e']['i'] == 13


def test_set_save(db):
    _id = db.db.testdoc.insert({'e': {'i': 13}})
    doc = db.get_queryset(Doc).find_one(_id)

    doc.e.i = 26
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data == {'_id': _id, 'e': {'i': 26}}


def test_delete_save(db):
    _id = db.db.testdoc.insert({'e': {'i': 13}})
    doc = db.get_queryset(Doc).find_one(_id)

    del doc.e
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data == {'_id': _id}


def test_delete_deep_save(db):
    _id = db.db.testdoc.insert({'e': {'i': 13, 'id': ObjectId()}})
    doc = db.get_queryset(Doc).find_one(_id)

    del doc.e.i
    db.save(doc)

    data = db.db.testdoc.find_one({'_id': _id})
    assert data == {'_id': _id, 'e': {'id': doc.e.id}}


def test_id_load_default():
    doc = Doc({'e': {'i': 13}})
    assert hasattr(doc, 'e')
    assert hasattr(doc.e, 'i')
    assert hasattr(doc.e, 'id')
    assert doc.e.i == 13
    assert isinstance(doc.e.id, ObjectId)
    assert doc.e.id is doc.e.id


def test_id_insert(db):
    doc = Doc({'e': {'i': 13}})

    # eid = doc.e.id
    db.insert(doc)
    # assert doc.e.id == eid
    # db.reload(doc)

    data = db.db.testdoc.find_one({'_id': doc.id})
    assert data['e']['id'] == doc.e.id


def test_fake():
    doc = create_fake(Doc)
    assert isinstance(doc.e, EDoc)
    assert isinstance(doc.e.i, int)
    assert isinstance(doc.e.s, str)
    assert isinstance(doc.e.e, EDoc)



from yadm.fields.base import pass_null
from yadm.fields.simple import StringField


class InvalidEmail(ValueError):
    """ Raise if value is not correct email address
    """


class EmailField(StringField):
    def get_fake(self, document, faker, depth):
        return faker.email().lower()

    @pass_null
    def prepare_value(self, document, value):
        value = super().prepare_value(document, value).lower()
        self.check_email(value)
        return value

    @classmethod
    def check_email(cls, value):
        """ Check email classmethod.

        Raise `ValueError` if value is not email.

        Usage: `EmailField.check_email(some_untrusted_email)`
        """
        if '@' not in value:
            cls._raise(value)

        parts = value.split('@')
        if len(parts) != 2 or len([p for p in parts if p]) != 2:
            cls._raise(value)

    @staticmethod
    def _raise(value):
        raise InvalidEmail('"{}" is not email'.format(value))



from unittest import SkipTest

import pytest

from yadm import fields
from yadm.documents import Document, EmbeddedDocument


class EDoc(EmbeddedDocument):
    i = fields.IntegerField()
    s = fields.StringField()


class Doc(Document):
    __collection__ = 'docs'
    li = fields.ListField(fields.EmbeddedDocumentField(EDoc))


def test_save(db):
    doc = Doc()

    edoc = EDoc()
    edoc.i = 13
    doc.li.append(edoc)

    assert edoc.__parent__ is doc.li

    edoc = EDoc()
    edoc.i = 42
    doc.li.append(edoc)

    db.insert(doc)

    data = db.db.docs.find_one()

    assert 'li' in data
    assert len(data['li']) == 2
    assert 'i' in data['li'][0]
    assert data['li'][0]['i'] == 13
    assert data['li'][1]['i'] == 42


def test_load(db):
    db.db.docs.insert({'li': [{'i': 13}, {'i': 42}]})

    doc = db.get_queryset(Doc).find_one()

    assert hasattr(doc, 'li')
    assert len(doc.li) == 2
    assert doc.li[0].__parent__ is doc.li
    assert doc.li[1].__parent__ is doc.li

    for item in doc.li:
        assert isinstance(item, EDoc)

    assert doc.li[0].i == 13
    assert doc.li[1].i == 42


def test_push_reload(db):
    doc = Doc()
    db.insert(doc)

    edoc = EDoc()
    edoc.i = 13
    doc.li.push(edoc)

    db.db.docs.update({}, {'$push': {'li': {'i': 32}}})
    assert len(doc.li) == 1

    edoc = EDoc()
    edoc.i = 42
    doc.li.push(edoc)

    assert len(doc.li) == 3
    assert doc.li[0].i == 13
    assert doc.li[1].i == 32
    assert doc.li[2].i == 42

    data = db.db.docs.find_one()

    assert data['li'][0]['i'] == 13
    assert data['li'][1]['i'] == 32
    assert data['li'][2]['i'] == 42


def test_push_no_reload(db):
    doc = Doc()
    db.insert(doc)

    edoc = EDoc()
    edoc.i = 13
    doc.li.push(edoc)

    db.db.docs.update({}, {'$push': {'li': {'i': 32}}})
    assert len(doc.li) == 1

    edoc = EDoc()
    edoc.i = 42
    doc.li.push(edoc, reload=False)

    assert len(doc.li) == 2
    assert doc.li[0].i == 13
    assert doc.li[1].i == 42

    data = db.db.docs.find_one()

    assert data['li'][0]['i'] == 13
    assert data['li'][1]['i'] == 32
    assert data['li'][2]['i'] == 42


def test_replace(db):
    doc = Doc()
    doc.li.append(EDoc(i=13, s='13'))
    doc.li.append(EDoc(i=42, s='42'))
    db.insert(doc)

    doc.li.replace({'i': 13}, EDoc(i=26))

    data = db.db.docs.find_one()

    assert 'li' in data
    assert len(data['li']) == 2
    assert 'i' in data['li'][0]
    assert data['li'][0]['i'] == 26
    assert 's' not in data['li'][0]
    assert data['li'][1]['i'] == 42
    assert data['li'][1]['s'] == '42'


def test_update(db):
    doc = Doc()
    doc.li.append(EDoc(i=13, s='13'))
    doc.li.append(EDoc(i=42, s='42'))
    db.insert(doc)

    doc.li.update({'i': 13}, {'i': 26})

    data = db.db.docs.find_one()

    assert 'li' in data
    assert len(data['li']) == 2
    assert 'i' in data['li'][0]
    assert data['li'][0]['i'] == 26
    assert data['li'][0]['s'] == '13'
    assert data['li'][1]['i'] == 42
    assert data['li'][1]['s'] == '42'


class DeepEEDoc(EmbeddedDocument):
    i = fields.IntegerField()
    s = fields.StringField()


class DeepEDoc(EmbeddedDocument):
    lie = fields.ListField(fields.EmbeddedDocumentField(DeepEEDoc))


class DeepDoc(Document):
    __collection__ = 'docs'
    li = fields.ListField(fields.EmbeddedDocumentField(DeepEDoc))


def test_deep_save(db):
    doc = DeepDoc()
    doc.li.append(DeepEDoc())

    eedoc = DeepEEDoc()
    eedoc.i = 13
    doc.li[0].lie.append(eedoc)

    db.save(doc)

    data = db.db.docs.find_one()

    assert 'li' in data
    assert len(data['li']) == 1
    assert 'lie' in data['li'][0]
    assert len(data['li'][0]['lie']) == 1
    assert 'i' in data['li'][0]['lie'][0]
    assert data['li'][0]['lie'][0]['i'] == 13


def test_deep_load(db):
    db.db.docs.insert({'li': [{'lie': [{'i': 13}]}]})

    doc = db.get_queryset(DeepDoc).find_one()

    assert hasattr(doc, 'li')
    assert len(doc.li) == 1
    assert hasattr(doc.li[0], 'lie')
    assert len(doc.li[0].lie) == 1
    assert hasattr(doc.li[0].lie[0], 'i')
    assert doc.li[0].lie[0].i == 13


def test_deep_push(db):
    doc = DeepDoc()
    db.insert(doc)

    doc.li.push(DeepEDoc())

    eedoc = DeepEEDoc()
    eedoc.i = 13
    doc.li[0].lie.push(eedoc, reload=False)
    db.reload(doc)

    data = db.db.docs.find_one()

    assert 'li' in data
    assert len(data['li']) == 1
    assert 'lie' in data['li'][0]
    assert len(data['li'][0]['lie']) == 1
    assert 'i' in data['li'][0]['lie'][0]
    assert data['li'][0]['lie'][0]['i'] == 13


def test_deep_push_valueerror(db):
    doc = DeepDoc()
    db.insert(doc)

    doc.li.push(DeepEDoc())

    eedoc = DeepEEDoc()
    eedoc.i = 13

    with pytest.raises(ValueError):
        doc.li[0].lie.push(eedoc)


def test_deep_pull(db):
    _id = db.db.docs.insert({'li': [{'lie': [{'i': 13}, {'i': 42}]}]})
    doc = db.get_queryset(DeepDoc).find_one({'_id': _id})

    doc.li[0].lie.pull({'i': 42}, reload=False)
    db.reload(doc)

    data = db.db.docs.find_one({'_id': _id})

    assert 'li' in data
    assert len(data['li']) == 1
    assert 'lie' in data['li'][0]
    assert len(data['li'][0]['lie']) == 1
    assert 'i' in data['li'][0]['lie'][0]
    assert data['li'][0]['lie'][0]['i'] == 13



"""
Work with embedded documents.

.. code-block:: python

    class EDoc(EmbeddedDocument):
        i = fields.IntegerField()

    class Doc(Document):
        __collection__ = 'docs'
        edoc = EmbeddedDocumentField(EDoc)

    doc = Doc()
    doc.edoc = EDoc()
    doc.edoc.i = 13
    db.insert(doc)
"""
import random

from yadm.common import EnclosedDocDescriptor
from yadm.documents import EmbeddedDocument
from yadm.markers import AttributeNotSet
from yadm.fields.base import Field, pass_null
from yadm.serialize import to_mongo, from_mongo
from yadm.testing import create_fake


class BaseEmbeddedDocumentField(Field):
    def get_embedded_document_class(self, document, value):
        """ Return class of embedded document for field.
        """
        raise NotImplementedError()

    @pass_null
    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return value

        elif isinstance(value, EmbeddedDocument):
            value.__parent__ = document
            value.__name__ = self.name

        else:
            raise TypeError("Only EmbeddedDocument is allowed, but {!r} given"
                            "".format(type(value)))

        return value

    @pass_null
    def to_mongo(self, document, value):
        return to_mongo(value)

    @pass_null
    def from_mongo(self, document, value):
        ed_class = self.get_embedded_document_class(document, value)
        return from_mongo(ed_class, value, parent=document, name=self.name)


class EmbeddedDocumentField(BaseEmbeddedDocumentField):
    """ Field for embedded objects.

    :param EmbeddedDocument embedded_document_class:
        class for embedded document
    :param bool auto_create: automatic creation embedded
        document from access
    """
    auto_create = True
    embedded_document_class = EnclosedDocDescriptor('embedded')

    def __init__(self, embedded_document_class, *,
                 auto_create=True, **kwargs):
        super().__init__(**kwargs)
        self.embedded_document_class = embedded_document_class
        self.auto_create = auto_create

    def get_embedded_document_class(self, document=None, value=None):
        return self.embedded_document_class

    def get_if_attribute_not_set(self, document):
        """ Call if key not exist in document.

        If auto_create is True, create and return new
        embedded document. Else AttributeError is raised.
        """
        if self.auto_create:
            ed_class = self.get_embedded_document_class(document)
            return ed_class(__parent__=document, __name__=self.name)
        else:
            return super().get_if_attribute_not_set(document)

    def get_fake(self, document, faker, depth):
        return create_fake(
            self.get_embedded_document_class(document),
            __parent__=document,
            __name__=self.name,
            __faker__=faker,
            __depth__=depth,
        )

    @pass_null
    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return value

        ed_class = self.get_embedded_document_class(document, value)

        if isinstance(value, dict):
            value = ed_class(__parent__=document, __name__=self.name, **value)

        elif isinstance(value, ed_class):
            value.__parent__ = document
            value.__name__ = self.name

        else:
            raise TypeError("Only {!r}, dict or None is allowed, but {!r} given"
                            "".format(ed_class, type(value)))

        return value

    def copy(self):
        """ Return copy of field.
        """
        ed_class = self.get_embedded_document_class()
        return self.__class__(ed_class, smart_null=self.smart_null)


class TypedEmbeddedDocumentField(BaseEmbeddedDocumentField):
    """ Field for embedded document with variable types.

    :param str type_field: name of field in embedded document
        for select type
    :param dict types: map of type names to embedded document classes
    """
    type_field = None
    types = None

    def __init__(self, type_field=None, types=None, **kwargs):
        super().__init__(**kwargs)

        self.type_field = type_field or self.type_field
        self.types = types or self.types

        if self.types is None:
            raise TypeError("type attribute is not set")
        elif self.type_field is None:
            raise TypeError("type_field attribute is not set")

    def get_embedded_document_class(self, document, value):
        type_name = value.get(self.type_field, AttributeNotSet)
        ed_class = self.types.get(type_name, None)

        if ed_class is None:
            raise ValueError("Not found document type for {!r}"
                             "".format(type_name))
        else:
            return ed_class

    def get_fake(self, document, faker, depth):
        type_name = random.choice(list(self.types))
        ed_class = self.get_embedded_document_class(
            document=document,
            value={self.type_field: type_name},
        )

        return create_fake(
            ed_class,
            __parent__=document,
            __name__=self.name,
            __faker__=faker,
            __depth__=depth,
            **{self.type_field: type_name}
        )


class SimpleEmbeddedDocumentField(EmbeddedDocumentField):
    """ Field for simply create embedded documents.

    Usage:

        class Doc(Document):
            embedded = SimpleEmbeddedDocumentField({
                'i': IntegerField(),
                's': StringField(),
            })
    """
    embedded_document_class = None

    def __init__(self, fields, *, auto_create=True, **kwargs):
        if not isinstance(fields, dict):
            raise TypeError("First argument must be a dict, not {}"
                            "".format(type(fields)))
        elif not fields:
            raise ValueError("fields is empty")

        self.fields = fields

        super().__init__(None, auto_create=auto_create, **kwargs)

    def contribute_to_class(self, document_class, name):
        super().contribute_to_class(document_class, name)

        self.embedded_document_class = type(
            '{}__{}'.format(document_class.__name__, name),
            (EmbeddedDocument,),
            self.fields,
        )



from datetime import datetime, date, timedelta

import dateutil.parser
import pytz

from yadm.fields.base import DefaultMixin, Field, pass_null


class DatetimeField(DefaultMixin, Field):
    """ Field for time stamp.

    :param bool auto_now: datetime.now as default
        (default: False)
    """
    def __init__(self, *, auto_now=False, **kwargs):
        self.auto_now = auto_now
        super().__init__(**kwargs)

    @staticmethod
    def _fix_timezone(value):
        if value.tzinfo is None:
            return value.replace(tzinfo=pytz.utc)
        else:
            return value

    def get_default(self, document):
        if self.auto_now:
            return datetime.now(pytz.utc)
        else:
            return super().default

    def get_fake(self, document, faker, depth):
        if self.auto_now:
            return datetime.now(pytz.utc)
        else:
            return faker.date_time()

    @classmethod
    @pass_null
    def prepare_value(cls, document, value):
        if isinstance(value, datetime):
            return cls._fix_timezone(value)

        elif isinstance(value, date):
            return datetime(
                *value.timetuple()[:3],
                tz=pytz.utc
            )

        elif isinstance(value, str):
            value = dateutil.parser.parse(value)
            return cls._fix_timezone(value)

        else:
            raise TypeError("First value must be datetime,"
                            " date or string, but {!r}".format(type(value)))

    @classmethod
    @pass_null
    def to_mongo(cls, document, value):
        return cls._fix_timezone(value)

    @classmethod
    @pass_null
    def from_mongo(cls, document, value):
        return cls._fix_timezone(value)


class TimedeltaField(DefaultMixin, Field):
    def get_fake(self, document, faker, depth):
        return faker.time_delta()

    @classmethod
    @pass_null
    def prepare_value(cls, document, value):
        if isinstance(value, timedelta):
            return value
        else:
            raise TypeError("Only timedelta is allowed, but {} given"
                            "".format(type(value)))

    @classmethod
    @pass_null
    def to_mongo(cls, document, value):
        return value.total_seconds()

    @classmethod
    @pass_null
    def from_mongo(cls, document, value):
        return timedelta(seconds=value)



import random

from decimal import Decimal, Context, getcontext

from yadm import fields
from yadm.documents import Document


def test_to_mongo():
    field = fields.DecimalField()

    data = field.to_mongo(None, Decimal('123.45'))
    assert set(data.keys()) == {'i', 'e'}
    assert data['i'] == 12345
    assert data['e'] == -2


def test_from_mongo():
    field = fields.DecimalField()

    dec = field.from_mongo(None, {'i': 12345, 'e': -2})
    assert dec == Decimal('123.45')


def test_integer_from_digits():
    for _ in range(100):
        digits = [random.randint(0, 9) for _ in range(random.randint(5, 20))]
        integer = fields.DecimalField._integer_from_digits(digits)

    assert integer == int(''.join(str(i) for i in digits))


def test_func_int():
    field = fields.DecimalField()
    dec = field.prepare_value(None, 13)
    assert isinstance(dec, Decimal)
    assert dec == Decimal(13)


def test_func_str():
    field = fields.DecimalField()
    dec = field.prepare_value(None, '3.14')
    assert isinstance(dec, Decimal)
    assert dec == Decimal('3.14')


def test_context_default():
    field = fields.DecimalField()
    assert isinstance(field.context, Context)
    assert field.context == getcontext()


def test_context_init():
    context = Context(prec=5)
    field = fields.DecimalField(context=context)
    assert isinstance(field.context, Context)
    assert field.context == context


class Doc(Document):
    __collection__ = 'testdocs'
    dec = fields.DecimalField()


def test_save(db):
    doc = Doc()
    doc.dec = Decimal('3.14')
    db.save(doc)

    data = db.db.testdocs.find_one()

    assert 'dec' in data
    assert set(data['dec']) == {'i', 'e'}
    assert data['dec']['i'] == 314
    assert data['dec']['e'] == -2


def test_load(db):
    db.db.testdocs.insert({'dec': {'i': 314, 'e': -2}})

    doc = db.get_queryset(Doc).find_one()

    assert hasattr(doc, 'dec')
    assert isinstance(doc.dec, Decimal)
    assert doc.dec == Decimal('3.14')



import pytest
from bson import ObjectId

from yadm.documents import Document
from yadm.markers import AttributeNotSet
from yadm import fields


class Doc(Document):
    __collection__ = 'testdocs'
    i = fields.IntegerField()
    b = fields.BooleanField()


@pytest.fixture
def doc():
    return Doc()


def test__db(doc):
    assert doc.__db__ is None


def test_fields():
    assert set(Doc.__fields__) == {'_id', 'i', 'b'}


def test_inheritance_fields():
    class InhDoc(Doc):
        d = fields.DecimalField

    assert set(InhDoc.__fields__) == {'_id', 'i', 'b', 'd'}
    assert set(Doc.__fields__) == {'_id', 'i', 'b'}
    assert Doc.__fields__['i'] is not InhDoc.__fields__['i']
    assert Doc.__fields__['i'].document_class is Doc
    assert InhDoc.__fields__['i'].document_class is InhDoc


def test_changed(doc):
    assert not doc.__raw__
    assert not doc.__cache__
    assert not doc.__changed__

    doc.i = 13
    assert not doc.__raw__
    assert not doc.__cache__
    assert doc.__changed__ == {'i': 13}

    doc.b = True
    assert not doc.__raw__
    assert not doc.__cache__
    assert doc.__changed__ == {'i': 13, 'b': True}


def test_changed_clear(doc):
    doc.i = 13
    doc.__changed_clear__()
    assert not doc.__raw__
    assert doc.__cache__ == {'i': 13}
    assert not doc.__changed__


def test_raw_cache_changed(db, doc):
    assert not doc.__raw__
    assert not doc.__cache__
    assert not doc.__changed__

    doc.i = 13

    assert not doc.__raw__
    assert not doc.__cache__
    assert doc.__changed__ == {'i': 13}

    db.save(doc)

    assert not doc.__raw__
    assert doc.__cache__ == {'i': 13, '_id': doc.id, 'b': AttributeNotSet}
    assert not doc.__changed__

    doc = db.reload(doc, new_instance=True)

    assert doc.__raw__ == {'i': 13, '_id': doc.id, 'b': AttributeNotSet}
    assert doc.__cache__ == {'_id': doc.id}
    assert not doc.__changed__

    doc.b = False

    assert doc.__raw__ == {'i': 13, '_id': doc.id, 'b': AttributeNotSet}
    assert doc.__cache__ == {'_id': doc.id}
    assert doc.__changed__ == {'b': False}

    assert doc.i == 13  # call descriptor's get

    assert doc.__raw__ == {'i': 13, '_id': doc.id, 'b': AttributeNotSet}
    assert doc.__cache__ == {'i': 13, '_id': doc.id}
    assert doc.__changed__ == {'b': False}

    doc.i = 12  # call descriptor's set, for cover

    assert doc.__raw__ == {'i': 13, '_id': doc.id, 'b': AttributeNotSet}
    assert doc.__cache__ == {'i': 13, '_id': doc.id}
    assert doc.__changed__ == {'i': 12, 'b': False}


def test_eq():
    doc_a = Doc()
    doc_a.id = ObjectId()

    doc_b = Doc()
    doc_b.id = ObjectId()

    doc_c = Doc()
    doc_c.id = doc_a.id

    assert doc_a != doc_b
    assert doc_a == doc_c


def test_id(doc):
    with pytest.raises(AttributeError):
        doc.id


def test_attributeerror(db):
    _id = db.db.testdocs.insert({'i': 13})
    doc = db(Doc).find_one(_id)

    assert doc.i == 13

    with pytest.raises(AttributeError):
        doc.b


def test_notloaded(db):
    _id = db.db.testdocs.insert({'i': 13, 'b': False})
    doc = db(Doc).fields('i').find_one(_id)

    assert doc.i == 13

    with pytest.raises(fields.NotLoadedError):
        doc.b
